/*
 *
 * Copyright (c) 2016,2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Rafal Pietruch <r.pietruch@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-arm.c
 * @brief unwinding call stacks, functions specific for ARM
 */
#include "crash-stack.h"

#include <string.h>
#include <sys/user.h>

// definitions copied from previously used McTernan unwinder
#define REGS_REGULAR_NUM 13
#define REG_FP 11
#define REG_IP 12
#define REG_SP 13
#define REG_LR 14
#define REG_PC 15
#define REG_SPSR 16

/**
 * @brief Important registers for unwinding stack on ARM
 */
struct Regs {
	Dwarf_Addr regs[REGS_REGULAR_NUM];		///< regular registers
	Dwarf_Addr sp;							///< register: stack pointer
	Dwarf_Addr lr;							///< register: link register
	Dwarf_Addr pc;							///< register: program counter
	Dwarf_Addr spsr;						///< register: status register
};

typedef struct Regs Regs;					///< convenience type definition
static Regs g_regs;							///< static storage for target register values

static struct user_regs g_registers;	///< static storage for registers

void *_crash_stack_get_memory_for_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	if (strcmp(regname, "pc") == 0 || REG_PC == regnum)
		return &g_regs.pc;
	else if (strcmp(regname, "sp") == 0 || REG_SP == regnum)
		return &g_regs.sp;
	else if (strcmp(regname, "lr") == 0 || REG_LR == regnum)
		return &g_regs.lr;
	else if (strcmp(regname, "spsr") == 0 || REG_SPSR == regnum)
		return &g_regs.spsr;
	else if (regnum < REGS_REGULAR_NUM)
		return &g_regs.regs[regnum];
	return NULL;
}

void _crash_stack_set_registers(void *regbuf)
{
	struct user_regs *registers = regbuf;
	int i;
	for (i = 0; i < sizeof(registers->uregs)/sizeof(registers->uregs[0]); i++) {
		void *regmem = _get_place_for_register_value("", i);
		if (NULL != regmem)
			memcpy(regmem, &registers->uregs[i], sizeof(registers->uregs[i]));
	}
}

unsigned long _get_register_value(int n)
{
	return g_registers.uregs[n];
}

void _crash_stack_print_regs(FILE* outputfile)
{
	fprintf(outputfile, "\nRegister Information\n");
	fprintf(outputfile, "r0   = 0x%08lx, r1   = 0x%08lx\nr2   = 0x%08lx, r3   = 0x%08lx\n",
	       g_registers.uregs[0], g_registers.uregs[1],
	       g_registers.uregs[2], g_registers.uregs[3]);
	fprintf(outputfile, "r4   = 0x%08lx, r5   = 0x%08lx\nr6   = 0x%08lx, r7   = 0x%08lx\n",
	       g_registers.uregs[4], g_registers.uregs[5],
	       g_registers.uregs[6], g_registers.uregs[7]);
	fprintf(outputfile, "r8   = 0x%08lx, r9   = 0x%08lx\nr10  = 0x%08lx, fp   = 0x%08lx\n",
	       g_registers.uregs[8], g_registers.uregs[9],
	       g_registers.uregs[10], g_registers.uregs[REG_FP]);
	fprintf(outputfile, "ip   = 0x%08lx, sp   = 0x%08lx\nlr   = 0x%08lx, pc   = 0x%08lx\n",
	       g_registers.uregs[REG_IP], g_registers.uregs[REG_SP],
	       g_registers.uregs[REG_LR], g_registers.uregs[REG_PC]);
	/* XXX the label should probably write "spsr" */
	fprintf(outputfile, "cpsr = 0x%08lx\n", g_registers.uregs[REG_SPSR]);
}
