/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef NOTE_HPP__
#define NOTE_HPP__

#include "helpers.hpp"
#include "maps.hpp"

#include <string.h>
#include <sys/procfs.h>

#include <fstream>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#if defined(__arm__)
typedef user_regs user_regs_struct;
#define MACHINE EM_ARM
#elif defined(__aarch64__)
#define MACHINE EM_AARCH64
#elif defined(__x86_64__)
#define MACHINE EM_X86_64
#elif defined(__i386__)
#define MACHINE EM_386
#else
#error "architecture unsupported"
#endif

class Note {
 public:
	// man elf(5): name and desc starting address has a 4 byte alignment.
	static constexpr unsigned int PADDING = 4;
	uint32_t name_size;
	uint32_t desc_size;
	uint32_t type;
	std::string name;

	explicit Note(uint32_t type) : name_size(0), desc_size(0), type(type) {}
	virtual ~Note() {}

	virtual void SaveDataTo(std::ofstream &core_file) const {}
	virtual size_t GetSize() const {
		size_t size = 0;
		size += sizeof(name_size);
		size += name_size + CalculatePadding<Note::PADDING>(name_size);
		size += sizeof(desc_size);
		size += desc_size + CalculatePadding<Note::PADDING>(desc_size);
		size += sizeof(type);
		return size;
	}

	template <int to>
	static size_t CalculatePadding(size_t value) {
		static_assert((to & (to - 1)) == 0, "must be power of 2");
		return (to - value) & (to - 1);
	}

	void Pad(std::ofstream &core_file, size_t value) const {
		static const char b = 0;
		for (size_t i = 0; i < value; i++)
			core_file.write(&b, 1);
	}

	void SaveTo(std::ofstream &core_file) const {
		core_file.write(reinterpret_cast<const char*>(&name_size), sizeof(name_size));
		core_file.write(reinterpret_cast<const char*>(&desc_size), sizeof(desc_size));
		core_file.write(reinterpret_cast<const char*>(&type), sizeof(type));
		core_file.write(name.c_str(), name_size);
		Pad(core_file, CalculatePadding<Note::PADDING>(name_size));
		SaveDataTo(core_file);
	}
};

class NoteCorePRStatus : public Note {
 private:
	prstatus_t m_prstatus;
 public:
	NoteCorePRStatus(const user_regs_struct &regs, const pid_t pid) : Note(NT_PRSTATUS) {
		memset(&m_prstatus, 0, sizeof(m_prstatus));
		name = "CORE";
		name_size = name.length() + 1;
		desc_size = sizeof(m_prstatus);
		m_prstatus.pr_pid = pid;

		memcpy(&m_prstatus.pr_reg, &regs, sizeof(m_prstatus.pr_reg));
	}

	void SaveDataTo(std::ofstream &core_file) const {
		core_file.write(reinterpret_cast<const char*>(&m_prstatus), sizeof(m_prstatus));
		Pad(core_file, CalculatePadding<Note::PADDING>(desc_size));
	}
};

template <typename T>
class NoteCoreFile : public Note {
 private:
	typename T::uint count;
	typename T::uint page_size;
	std::vector<std::unique_ptr<typename T::Notefile>> files;
	std::vector<std::string> names;

 public:
	NoteCoreFile() : Note(NT_FILE), count(0), page_size(0) { }

	void ImportRecords(const std::vector<std::shared_ptr<MapRecord>> &records) {
		name = "CORE";
		name_size = name.length() + 1;
		desc_size = 2*sizeof(typename T::uint);

		for (const auto &record : records) {
			if (!fileExists(record->path))
				continue;
			desc_size += sizeof(typename T::Notefile);
			auto *entry = new typename T::Notefile();
			entry->start = record->start;
			entry->end = record->end;
			entry->file_ofs = record->offset;
			files.push_back(std::unique_ptr<typename T::Notefile>(entry));

			desc_size += record->path.length() + 1;
			names.push_back(record->path);
		}
	}

	void SaveDataTo(std::ofstream &core_file) const {
		typename T::uint cc = files.size();
		typename T::uint page_size = 1;

		core_file.write(reinterpret_cast<const char*>(&cc),
		                sizeof(typename T::uint));
		core_file.write(reinterpret_cast<const char*>(&page_size),
		                sizeof(typename T::uint));

		for (const auto &file : files)
			core_file.write(reinterpret_cast<const char*>(file.get()), sizeof(*file));

		for (const auto &name : names)
			core_file.write(reinterpret_cast<const char*>(name.c_str()), name.length() + 1);

		Pad(core_file, CalculatePadding<Note::PADDING>(desc_size));
	}
};

class NoteCoreAUXV : public Note {
 private:
	std::vector<char> m_auxv;

 public:
	NoteCoreAUXV() : Note(NT_AUXV) { }
	void SetAuxv(const std::vector<char> &auxv) {
		name = "CORE";

		m_auxv = auxv;

		name_size = name.length() + 1;
		desc_size = m_auxv.size();
	}
	void SaveDataTo(std::ofstream &core_file) const {
		core_file.write(m_auxv.data(), m_auxv.size());
	}

	const std::vector<char>& auxv() const {
		return m_auxv;
	}
};
#endif  // NOTE_HPP__
